"""Test the shaker driver responds with correct data."""
from random import uniform
from unittest import mock

import pytest

from ika import command_line
from ika.mock import Shaker

ADDRESS = 'fakeip:123'


@pytest.fixture
def driver():
    """Confirm the orbital shaker correctly initializes."""
    return Shaker(ADDRESS)


@pytest.fixture
def expected_info_response():
    """Return mocked data."""
    return {
            'name': "SHAKE AND BAKE",
            'version': 1,
            'software_ID': 2,
    }


@mock.patch('ika.Shaker', Shaker)
def test_driver_cli_with_info(capsys):
    """Confirm the commandline interface works."""
    command_line([ADDRESS, '--type', 'shaker'])
    captured = capsys.readouterr()
    assert 'speed' in captured.out
    assert 'name' in captured.out


@mock.patch('ika.Shaker', Shaker)
def test_driver_cli(capsys):
    """Confirm the commandline interface works with --no-info."""
    command_line([ADDRESS, '--type', 'shaker', '--no-info'])
    captured = capsys.readouterr()
    assert 'speed' in captured.out
    assert 'name' not in captured.out


async def test_get_response(driver, expected_info_response):
    """Confirm that the driver returns correct values on get_info() calls."""
    assert expected_info_response == await driver.get_info()


async def test_readme_example(expected_info_response):
    """Confirm the readme example using an async context manager works."""
    async def get():
        async with Shaker(ADDRESS) as device:
            response = await device.get()       # Get speed, torque, temp, setpoints
            assert 'speed' in response
            assert expected_info_response == await device.get_info()  # Get name
    await get()


async def test_setpoint_roundtrip():
    """Confirm that the setpoint can be updated."""
    async def get():
        async with Shaker(ADDRESS) as device:
            with pytest.raises(ValueError, match="Setpoint invalid"):
                await device.set(equipment='shaker', setpoint=299)
            with pytest.raises(ValueError, match="Setpoint invalid"):
                await device.set(equipment='shaker', setpoint=3001)
            with pytest.raises(ValueError, match="Setpoint invalid"):
                await device.set(equipment='heater', setpoint=0)
            with pytest.raises(ValueError, match="Setpoint invalid"):
                await device.set(equipment='heater', setpoint=101)
            speed_sp = round(uniform(300, 1000), 0)
            temp_sp = round(uniform(30, 100), 2)
            await device.set(equipment='shaker', setpoint=speed_sp)
            await device.set(equipment='heater', setpoint=temp_sp)
            response = await device.get()
            assert speed_sp == response['speed']['setpoint']
            assert temp_sp == response['temp']['setpoint']
    await get()


async def test_start_stop():
    """Confirm that the shaker motor can be controlled."""
    async def get():
        async with Shaker(ADDRESS) as device:
            response = await device.get()
            assert response['temp']['active'] is False
            assert response['speed']['active']

            await device.control(equipment='heater', on=True)
            response = await device.get()
            assert response['temp']['active']
            assert response['speed']['active']

            await device.control(equipment='shaker', on=False)
            response = await device.get()
            assert response['temp']['active']
            assert response['speed']['active'] is False

            await device.control(equipment='heater', on=False)
            await device.control(equipment='shaker', on=True)
            response = await device.get()
            assert response['temp']['active'] is False
            assert response['speed']['active']
    await get()



"""Test the hotplate driver responds with correct data."""
from random import uniform
from unittest import mock

import pytest

from ika import command_line
from ika.mock import Hotplate

ADDRESS = 'fakeip:123'


@pytest.fixture
def driver():
    """Confirm the hotplate correctly initializes."""
    return Hotplate(ADDRESS)


@pytest.fixture
def expected_info_response():
    """Return mocked hotplate data."""
    return {
        "name": "SPINNY HOT THING",
        "temp_limit": 150.0,
        "device_type": 1,
    }


@mock.patch('ika.Hotplate', Hotplate)
def test_driver_cli_with_info(capsys):
    """Confirm the commandline interface works."""
    command_line([ADDRESS, '--type', 'hotplate'])
    captured = capsys.readouterr()
    assert "temp_limit" in captured.out
    assert "name" in captured.out


@mock.patch('ika.Hotplate', Hotplate)
def test_driver_cli(capsys):
    """Confirm the commandline interface works with --no-info."""
    command_line([ADDRESS, '--type', 'hotplate', '--no-info'])
    captured = capsys.readouterr()
    assert "temp" in captured.out
    assert "name" not in captured.out


async def test_get_response(driver, expected_info_response):
    """Confirm that the driver returns correct values on get_info() calls."""
    assert expected_info_response == await driver.get_info()


async def test_readme_example(expected_info_response):
    """Confirm the readme example using an async context manager works."""
    async def get():
        async with Hotplate(ADDRESS) as device:
            response = await device.get()       # Get speed, torque, temp, setpoints
            assert "process_temp" in response
            assert expected_info_response == await device.get_info()  # Get name
    await get()


async def test_setpoint_roundtrip():
    """Confirm that setpoints can be updated."""
    async def get():
        async with Hotplate(ADDRESS, include_surface_control=True) as device:
            process_sp = round(uniform(15, 100), 2)
            surface_sp = round(uniform(30, 150), 2)
            await device.set(equipment='process', setpoint=process_sp)
            await device.set(equipment='surface', setpoint=surface_sp)
            response = await device.get()
            assert process_sp == response['process_temp']['setpoint']
            assert surface_sp == response['surface_temp']['setpoint']
    await get()


async def test_start_stop():
    """Confirm that the heater and shaker motor can be controlled."""
    async def get():
        async with Hotplate(ADDRESS) as device:
            response = await device.get()
            assert response['process_temp']['active'] is False
            assert response['speed']['active'] is False

            await device.control(equipment='heater', on=True)
            response = await device.get()
            assert response['process_temp']['active']
            assert response['speed']['active'] is False

            await device.control(equipment='motor', on=True)
            response = await device.get()
            assert response['process_temp']['active']
            assert response['speed']['active']

            await device.control(equipment='heater', on=False)
            await device.control(equipment='motor', on=False)
            response = await device.get()
            assert response['process_temp']['active'] is False
            assert response['speed']['active'] is False
    await get()



"""Contains mocks for driver objects for offline testing."""
from __future__ import annotations

import asyncio
from random import uniform
from typing import Any
from unittest.mock import MagicMock

from .driver import Hotplate as RealHotplate
from .driver import OverheadStirrer as RealOverheadStirrer
from .driver import Shaker as RealShaker
from .driver import Vacuum as RealVacuum
from .driver import VacuumProtocol


class AsyncClientMock(MagicMock):
    """Magic mock that works with async methods."""

    async def __call__(self, *args, **kwargs):
        """Convert regular mocks into into an async coroutine."""
        return super().__call__(*args, **kwargs)


class OverheadStirrer(RealOverheadStirrer):
    """Mocks the overhead stirrer driver for offline testing."""

    def __init__(self, *args, **kwargs):
        """Set up connection parameters with default port."""
        super().__init__(*args, **kwargs)
        self.hw = AsyncClientMock()
        self.state: dict[str, Any] = {
            'name': 'STIRR GO WHIRRR',
            'torque_limit': 60.0,
            'speed_limit': 2000.0,
            'speed': {
                'setpoint': 0.0,
                'actual': 0.0,
                'active': False,
            },
            'torque': 0.0,
            'temp': 0.0,
        }

    async def query(self, command):
        """Return mock requests to queries."""
        if not self.lock:
            self.lock = asyncio.Lock()
        async with self.lock:  # lock releases on CancelledError
            if command == self.READ_DEVICE_NAME:
                return self.state['name']
            elif command == self.READ_TORQUE_LIMIT:
                return self.state['torque_limit']
            elif command == self.READ_SPEED_LIMIT:
                return self.state['speed_limit']
            elif command == self.READ_ACTUAL_SPEED:
                return round(uniform(30, 120), 2)
            elif command == self.READ_ACTUAL_TORQUE:
                return round(uniform(0, 10), 2)
            elif command == self.READ_PT1000:
                return round(uniform(15, 60), 2)
            elif command == self.READ_MOTOR_STATUS:
                return self.state['speed']['active']
            elif command == self.READ_SET_SPEED:
                return self.state['speed']['setpoint']
            elif command == self.START_MOTOR:
                self.state['speed']['active'] = True
                return self.START_MOTOR
            elif command == self.STOP_MOTOR:
                self.state['speed']['active'] = False
                return self.STOP_MOTOR

    async def command(self, command):
        """Update mock state with commands."""
        if not self.lock:
            self.lock = asyncio.Lock()
        async with self.lock:  # lock releases on CancelledError
            command, value = command.split(" ")
            if command == self.SET_SPEED.strip():
                self.state['speed']['setpoint'] = float(value)
            elif command == self.SET_SPEED_LIMIT.strip():
                self.state['speed_limit'] = float(value)
            elif command == self.SET_TORQUE_LIMIT.strip():
                self.state['torque_limit'] = float(value)


class Hotplate(RealHotplate):
    """Mocks the hotplate driver for offline testing."""

    def __init__(self, *args, **kwargs):
        """Set up connection parameters with default port."""
        super().__init__(*args, **kwargs)
        self.hw = AsyncClientMock()
        self.state: dict[str, dict[str, bool | float | str]] = {
            "info": {
                "name": "SPINNY HOT THING",
                "device_type": 1,
                "temp_limit": 150.0,
            },
            "process_temp": {
                "setpoint": 50,
                "active": False,
            },
            "surface_temp": {
                "setpoint": 70,
                "active": False,
            },
            "speed": {
                "setpoint": 300,
                "active": False,
            },
            "fluid_temp": {
                "actual": 100,
            }
        }

    async def query(self, command):
        """Return mock requests to queries."""
        await asyncio.sleep(uniform(0.0, 0.05))
        if not self.lock:
            self.lock = asyncio.Lock()
        async with self.lock:  # lock releases on CancelledError
            if command == self.READ_DEVICE_NAME:
                return self.state["info"]["name"]
            elif command == self.READ_DEVICE_TYPE:
                return self.state["info"]["device_type"]
            elif command == self.READ_TEMP_LIMIT:
                return self.state["info"]["temp_limit"]
            elif command == self.READ_ACTUAL_SPEED:
                return round(uniform(10, 100), 2)
            elif command == self.READ_ACTUAL_PROCESS_TEMP:
                return round(uniform(15, 100), 2)
            elif command == self.READ_ACTUAL_SURFACE_TEMP:
                return round(uniform(80, 120), 2)
            elif command == self.READ_ACTUAL_FLUID_TEMP:
                return round(uniform(20, 110), 2)
            elif command == self.READ_SURFACE_TEMP_SETPOINT:
                return self.state["surface_temp"]["setpoint"]
            elif command == self.READ_PROCESS_TEMP_SETPOINT:
                return self.state["process_temp"]["setpoint"]
            elif command == self.READ_SPEED_SETPOINT:
                return self.state["speed"]["setpoint"]
            elif command == self.READ_SHAKER_STATUS:
                return self.state["speed"]["active"]
            elif command == self.READ_PROCESS_HEATER_STATUS:
                return self.state["process_temp"]["active"]
            elif command == self.READ_SURFACE_HEATER_STATUS:
                return self.state["surface_temp"]["active"]

    async def command(self, command):
        """Update mock state with commands."""
        if not self.lock:
            self.lock = asyncio.Lock()
        async with self.lock:  # lock releases on CancelledError
            if command == self.START_THE_MOTOR:
                self.state["speed"]["active"] = True
            elif command == self.STOP_THE_MOTOR:
                self.state["speed"]["active"] = False
            elif command == self.START_THE_HEATER:
                self.state["process_temp"]["active"] = True
            elif command == self.STOP_THE_HEATER:
                self.state["process_temp"]["active"] = False
            else:
                command, value = command.split(" ")
            if command == self.SET_PROCESS_TEMP_SETPOINT.strip():
                self.state["process_temp"]["setpoint"] = float(value)
            elif command == self.SET_SURFACE_TEMP_SETPOINT.strip():
                self.state["surface_temp"]["setpoint"] = float(value)


class Shaker(RealShaker):
    """Mocks the orbital shaker driver for offline testing."""

    def __init__(self, *args, **kwargs):
        """Set up connection parameters with default port."""
        super().__init__(*args, **kwargs)
        self.hw = AsyncClientMock()
        self.state: dict[str, Any] = {
            'name': "SHAKE AND BAKE",
            'temp': {
                'setpoint': 50,
                'actual': 25,
                'active': False,
            },
            'speed': {
                'setpoint': 100,
                'actual': 100,
                'active': True,
            },
            'version': 1,
            'software_ID': 2,
        }

    async def query(self, command):
        """Return mock requests to queries."""
        if not self.lock:
            self.lock = asyncio.Lock()
        async with self.lock:  # lock releases on CancelledError
            if command == self.READ_DEVICE_NAME:
                return self.state['name']
            elif command == self.READ_SET_TEMPERATURE:
                return self.state['temp']['setpoint']
            elif command == self.READ_ACTUAL_TEMPERATURE:
                return self.state['temp']['actual']
            elif command == self.READ_HEATER_STATUS:
                return self.state['temp']['active']
            elif command == self.READ_SET_SPEED:
                return self.state['speed']['setpoint']
            elif command == self.READ_ACTUAL_SPEED:
                return self.state['speed']['actual']
            elif command == self.READ_MOTOR_STATUS:
                return self.state['speed']['active']
            elif command == self.READ_SOFTWARE_VERSION:
                return self.state['version']
            elif command == self.READ_SOFTWARE_ID:
                return self.state['software_ID']

    async def command(self, command):
        """Update mock state with commands."""
        if not self.lock:
            self.lock = asyncio.Lock()
        async with self.lock:  # lock releases on CancelledError
            if command == self.START_MOTOR:
                self.state['speed']['active'] = True
            elif command == self.STOP_MOTOR:
                self.state['speed']['active'] = False
            elif command == self.START_HEATER:
                self.state['temp']['active'] = True
            elif command == self.STOP_HEATER:
                self.state['temp']['active'] = False
            else:
                command, value = command.split(" ")
            if command == self.SET_TEMP.strip():
                self.state['temp']['setpoint'] = float(value)
            elif command == self.SET_SPEED.strip():
                self.state['speed']['setpoint'] = float(value)


class Vacuum(RealVacuum):
    """Mocks the vacuum driver for offline testing."""

    def __init__(self, *args, **kwargs):
        """Set up connection parameters with default port."""
        super().__init__(*args, **kwargs)
        self.hw = AsyncClientMock()
        self.state: dict[str, Any] = {
            'name': 'THIS SUCKS',
            'active': False,
            'mode': VacuumProtocol.Mode.AUTOMATIC.name,
            'version': '1.3.001',
            'pressure': {
                'setpoint': 0.0,
                'actual': 0.0,
            }
        }

    async def query(self, command) -> str:
        """Return mock requests to queries."""
        if not self.lock:
            self.lock = asyncio.Lock()
        async with self.lock:  # lock releases on CancelledError
            if command == self.READ_DEVICE_NAME:
                return self.state['name']
            elif command == self.READ_SET_PRESSURE:
                return str(self.state['pressure']['setpoint'])
            elif command == self.READ_ACTUAL_PRESSURE:
                return str(self.state['pressure']['actual'])
            elif command == self.READ_VAC_STATUS:
                return '75' if not self.state['active'] else '12345'
            elif command == self.READ_SOFTWARE_VERSION:
                return self.state['version']
            elif command == self.READ_VAC_MODE:
                return VacuumProtocol.Mode[self.state['mode']].value
            elif command == self.START_MEASUREMENT:
                self.state['active'] = True
            elif command == self.STOP_MEASUREMENT:
                self.state['active'] = False
            else:
                command, value = command.split(" ", 1)
            if command == self.SET_PRESSURE.strip():
                self.state['pressure']['setpoint'] = float(value)
            elif command == self.SET_DEVICE_NAME.strip():
                self.state['name'] = value
                return ""
            elif command == self.SET_VAC_MODE.strip():
                self.state['mode'] = VacuumProtocol.Mode(value).name
            return command



"""IKA TCP adapter for overhead stirrers and hotplates."""

import asyncio
import logging
from abc import ABC
from enum import Enum
from typing import Any, Dict

from ika.util import Client, SerialClient, TcpClient

logger = logging.getLogger('ika')


class IKADevice(ABC):
    """Abstract base class for IKA devices."""

    def __init__(self, address, **kwargs):
        """Set up connection parameters, serial or IP address and port."""
        if address.startswith('/dev') or address.startswith('COM'):  # serial
            self.hw: Client = SerialClient(address=address, **kwargs)
        else:
            self.hw = TcpClient(address=address, **kwargs)
        self.lock = None  # needs to be initialized later, when the event loop exists

    async def __aenter__(self, *args):
        """Provide async enter to context manager."""
        return self

    async def __aexit__(self, *args):
        """Provide async exit to context manager."""
        return

    async def query(self, query) -> str:
        """Query the device and return its response."""
        if not self.lock:
            self.lock = asyncio.Lock()
        async with self.lock:  # lock releases on CancelledError
            return await self.hw._write_and_read(query)

    async def command(self, command) -> None:
        """Send a command to the device and don't expect a response."""
        if not self.lock:
            self.lock = asyncio.Lock()
        async with self.lock:  # lock releases on CancelledError
            await self.hw._write(command)

    async def reset(self) -> None:
        """Reset the device."""
        await self.command('RESET')


class OverheadStirrerProtocol:
    """Protocol for communicating with an overhead stirrer.

    Command syntax and format from the manual:
        - commands and parameters are transmitted as capital letters
        - commands and parameters including successive parameters are separated by at least
          one space (hex 0x20)
        - each individual command (including parameters and data and each response are terminated
          with Blank CR LF (hex 0x20 hex 0x0d hex 0x0A) and have a maximum length of 80 characters
        - the decimal separator in a number is a dt (hex 0x2E)
    """

    # overhead stirrer NAMUR commands
    READ_DEVICE_NAME = "IN_NAME"
    SET_DEVICE_NAME = "OUT_NAME "  # set name, undocumented
    READ_PT1000 = "IN_PV_3"  # read PT1000 value - temperature from the temperature sensor
    READ_ACTUAL_SPEED = "IN_PV_4"  # current actual speed
    READ_ACTUAL_TORQUE = "IN_PV_5"  # current actual torque
    READ_SET_SPEED = "IN_SP_4"  # speed setpoint
    READ_TORQUE_LIMIT = "IN_SP_5"
    READ_SPEED_LIMIT = "IN_SP_6"
    READ_SAFETY_SPEED = "IN_SP_8"  # safety speed value
    SET_SPEED = "OUT_SP_4 "  # set the speed setpoint
    SET_TORQUE_LIMIT = "OUT_SP_5 "  # set the torque limit
    SET_SPEED_LIMIT = "OUT_SP_6 "
    SET_SAFETY_SPEED = "OUT_SP_8 "
    START_MOTOR = "START_4"  # start stirring
    STOP_MOTOR = "STOP_4"  # stop stirring
    READ_MOTOR_STATUS = "STATUS_4"  # running status, undocumented in manual
    RESET = 'RESET'
    # todo change the direction or rotation with "OUT_MODE_n" (n = 1 or 2).
    # doesn't seem to work with the microstar C
    SET_ROTATION_CLOCKWISE = "OUT_MODE_1"
    SET_ROTATION_CLOCKWISE = "OUT_MODE_2"  # Fixme: verify
    READ_ROTATION_DIRECTION = "IN_MODE"  # todo doesn't seem to work with the microstar C


class OverheadStirrer(OverheadStirrerProtocol, IKADevice):
    """Driver for IKA overhead stirrer."""

    async def get(self):
        """Get overhead stirrer speed, torque, and external temperature reading."""
        speed = await self.query(self.READ_ACTUAL_SPEED)
        speed_sp = await self.query(self.READ_SET_SPEED)
        motor_status = await self.query(self.READ_MOTOR_STATUS)
        torque = await self.query(self.READ_ACTUAL_TORQUE)
        temp = await self.query(self.READ_PT1000)
        # FIXME handle case where temp probe is unplugged
        response = {
            'speed': {
                'setpoint': speed_sp,
                'actual': speed,
                'active': motor_status,
            },
            'torque': torque,
            'temp': temp,
        }
        return response

    async def get_info(self):
        """Get name and safety setpoints of overhead stirer."""
        name = await self.query(self.READ_DEVICE_NAME)
        torque_limit = await self.query(self.READ_TORQUE_LIMIT)
        speed_limit = await self.query(self.READ_SPEED_LIMIT)
        response = {
            'name': name,
            'torque_limit': torque_limit,
            'speed_limit': speed_limit,
        }
        return response

    async def set(self, equipment='speed', setpoint=0):
        """Set a parameter to the specified value."""
        if equipment == 'speed':
            await self.command(self.SET_SPEED + str(setpoint))
        elif equipment == 'speed_limit':
            await self.command(self.SET_SPEED_LIMIT + str(setpoint))
        elif equipment == 'torque_limit':
            await self.command(self.SET_TORQUE_LIMIT + str(setpoint))
        else:
            raise ValueError("Call with 'speed', 'speed_limit', or 'torque_limit'")

    async def control(self, on: bool):
        """Control the overhead stirrer motor."""
        await self.query(self.START_MOTOR if on else self.STOP_MOTOR)


class HotplateProtocol:
    """Protocol for communicating with an overhead stirrer.

    Command syntax and format from the manual:
        - commands and parameters are transmitted as capital letters
        - commands and parameters including successive parameters are separated by at least
          one space (hex 0x20)
        - each individual command (including parameters and data and each response are terminated
          with Blank CR LF (hex 0x20 hex 0x0d hex 0x0A) and have a maximum length of 80 characters
        - the decimal separator in a number is a dt (hex 0x2E)
    """

    # hotplate NAMUR commands
    READ_DEVICE_NAME = "IN_NAME"
    READ_DEVICE_TYPE = "IN_TYPE"
    READ_ACTUAL_PROCESS_TEMP = "IN_PV_1"
    READ_ACTUAL_SURFACE_TEMP = "IN_PV_2"
    READ_ACTUAL_SPEED = "IN_PV_4"
    READ_ACTUAL_FLUID_TEMP = "IN_PV_7"  # for double temp probe "PT1000"
    READ_VISCOSITY_TREND_VALUE = "IN_PV_5"
    READ_PROCESS_TEMP_SETPOINT = "IN_SP_1"
    READ_SURFACE_TEMP_SETPOINT = "IN_SP_2"
    READ_TEMP_LIMIT = "IN_SP_3"  # find the set safe temperature of the plate, the target/set
    # temperature the plate can go to is 50 degrees beneath this
    READ_SPEED_SETPOINT = "IN_SP_4"
    READ_PROCESS_HEATER_STATUS = "STATUS_1"  # undocumented in manual
    READ_SURFACE_HEATER_STATUS = "STATUS_2"  # undocumented in manual
    READ_SHAKER_STATUS = "STATUS_4"  # undocumented in manual
    SET_PROCESS_TEMP_SETPOINT = "OUT_SP_1 "  # requires a value to be appended
    SET_SURFACE_TEMP_SETPOINT = "OUT_SP_2 "
    SET_SPEED_SETPOINT = "OUT_SP_4 "  # requires a value to be appended
    START_THE_HEATER = "START_1"
    STOP_THE_HEATER = "STOP_1"
    START_THE_MOTOR = "START_4"
    STOP_THE_MOTOR = "STOP_4"
    RESET = "RESET"
    SET_OPERATING_MODE_A = "SET_MODE_A"
    SET_OPERATING_MODE_B = "SET_MODE_B"
    SET_OPERATING_MODE_D = "SET_MODE_D"
    SET_WD_SAFETY_LIMIT_TEMPERATURE_WITH_SET_VALUE_ECHO = "OUT_SP_12@"
    # requires a value to be appended
    SET_WD_SAFETY_LIMIT_SPEED_WITH_SET_VALUE_ECHO = "OUT_SP_42@"
    # requires a value to be appended
    WATCHDOG_MODE_1 = "OUT_WD1@"  # requires a watchdog time (20-1500 s) to be appended to the end
    # This command launches the watchdog function and must be transmitted within the set time.
    # In watchdog mode 1, if event WD1 occurs, the heating and stirring functions are switched off
    #  and ER 2 is displayed
    WATCHDOG_MODE_2 = "OUT_WD2@"  # requires a watchdog time (20-1500 s) to be appended to the end
    # This command launches the watchdog function and must be transmitted within the set time.
    # In watchdog mode 2, if event WD2 occurs, the speed and temperature setpoint are set to their
    # watchdog setpoints.
    # the WD2 event can be reset with the command "OUT_WD2@0", which also stops the watchdog
    COMMUNICATE_WITH_EUROSTAR = "IN_PV_4"
    # If this response is received, the hotplate has been erroneously configured to attempt to
    # communicate with a Eurostar overhead stirrer over RS-232.


class Hotplate(HotplateProtocol, IKADevice):
    """Driver for IKA hotplate stirrer."""

    def __init__(self, address, include_surface_control=False):
        """Set up connection parameters, IP address and port."""
        super().__init__(address)
        self.include_surface_control = include_surface_control

    async def get(self, include_surface_control=False):
        """Get hotplate speed, surface temperature, and process temperature readings."""
        speed = await self.query(self.READ_ACTUAL_SPEED)
        speed_sp = await self.query(self.READ_SPEED_SETPOINT)
        process_temp = await self.query(self.READ_ACTUAL_PROCESS_TEMP)
        process_temp_sp = await self.query(self.READ_PROCESS_TEMP_SETPOINT)
        fluid_temp = await self.query(self.READ_ACTUAL_FLUID_TEMP)
        shaker_status = await self.query(self.READ_SHAKER_STATUS)
        process_heater_status = await self.query(self.READ_PROCESS_HEATER_STATUS)
        surface_data = {
            'actual': await self.query(self.READ_ACTUAL_SURFACE_TEMP)
        }
        if self.include_surface_control:
            surface_data['setpoint'] = await self.query(self.READ_SURFACE_TEMP_SETPOINT)
            # surface_data['active'] = await self.query(self.READ_SURFACE_HEATER_STATUS)
            # FIXME figure out response value of '-90 02'
        # FIXME handle case where process temp probe is unplugged
        response = {
            'speed': {
                'setpoint': int(speed_sp) if isinstance(speed_sp, float) else speed_sp,
                'actual': int(speed) if isinstance(speed, float) else speed,
                'active': shaker_status,
            },
            'process_temp': {
                'setpoint': process_temp_sp,
                'actual': process_temp,
                'active': process_heater_status,
            },
            'surface_temp': surface_data,
            'fluid_temp': {
                'actual': fluid_temp,
            }
        }
        return response

    async def get_info(self):
        """Get name and safety setpoint of hotplate."""
        name = await self.query(self.READ_DEVICE_NAME)
        device_type = await self.query(self.READ_DEVICE_TYPE)
        temp_limit = await self.query(self.READ_TEMP_LIMIT)
        response = {
            'name': name,
            'device_type': device_type,
            'temp_limit': temp_limit,
        }
        return response

    async def control(self, equipment: str, on: bool):
        """Control the heater controlling process temperature, or shaker motor.

        Note: direct control of surface temperature is not implemented.
        """
        if equipment == 'heater':
            await self.command(self.START_THE_HEATER if on else self.STOP_THE_HEATER)
            # note - apparently after starting the heater it resets the setpoint to 0C
        elif equipment == 'motor':
            await self.command(self.START_THE_MOTOR if on else self.STOP_THE_MOTOR)
        else:
            raise ValueError(f'Equipment "{equipment} invalid. '
                             'Must be either "heater" or "motor"')

    async def set(self, equipment: str, setpoint: float):
        """Set a temperature or stirrer setpoint."""
        if equipment == 'process':
            await self.command(self.SET_PROCESS_TEMP_SETPOINT + str(setpoint))
        elif equipment == 'surface':
            await self.command(self.SET_SURFACE_TEMP_SETPOINT + str(setpoint))
        elif equipment == 'shaker':
            if setpoint < 50 or setpoint > 1700:
                raise ValueError(f"Cannot set shaker to {setpoint}RPM. "
                                 "Minimum shaker setpoint is 50RPM and maximum is 1700RPM.")
            # setpoints can be written as a decimal but the shaker will round off to int
            await self.command(self.SET_SPEED_SETPOINT + str(int(setpoint)))
        else:
            raise ValueError(f'Equipment "{equipment} invalid. '
                             'Must be "process", "surface", or "shaker"')

    async def reset(self):
        """Reset the hotplate, and turn off the heater and stirrer."""
        await self.command(self.RESET)


class ShakerProtocol:
    """Protocol for communicating with an orbital shaker.

    RS-232 Information
        - Transmission procedure: asynchronous character transmission in start-stop mode
        - Type of transmission: full duplex
        - 1 start bit; 7 character bits; 1 parity bit (even); 1 stop bit
        - Transmission speed: 9600 bit/s

    Command syntax and format from the manual:
        - commands and parameters are transmitted as capital letters
        - commands and parameters including successive parameters are separated by at least
          one space (hex 0x20)
        - each individual command (including parameters and data and each response are terminated
          with CR LF (hex 0x0d hex 0x0A) and have a maximum length of 80 characters
        - the decimal separator in a number is a dt (hex 0x2E)
    """

    # orbital shaker NAMUR commands
    READ_DEVICE_NAME = "IN_NAME"
    READ_ACTUAL_TEMPERATURE = "IN_PV_2"
    READ_ACTUAL_SPEED = "IN_PV_4"
    READ_SET_TEMPERATURE = "IN_SP_2"
    READ_SET_SPEED = "IN_SP_4"
    SET_TEMP = "OUT_SP_2 "
    SET_SPEED = "OUT_SP_4 "
    SET_WD_SAFETY_LIMIT_TEMPERATURE_WITH_SET_VALUE_ECHO = "OUT_SP_12@"
    # requires a value to be appended
    SET_WD_SAFETY_LIMIT_SPEED_WITH_SET_VALUE_ECHO = "OUT_SP_42@"
    # requires a value to be appended
    WATCHDOG_MODE_1 = "OUT_WD1@"  # requires a watchdog time (20-1500 s) to be appended to the end
    # This command launches the watchdog function and must be transmitted within the set time.
    # In watchdog mode 1, if event WD1 occurs, the heating and stirring functions are switched off
    #  and ER 2 is displayed
    WATCHDOG_MODE_2 = "OUT_WD2@"  # requires a watchdog time (20-1500 s) to be appended to the end
    # This command launches the watchdog function and must be transmitted within the set time.
    # In watchdog mode 2, if event WD2 occurs, the speed and temperature setpoint are set to their
    # watchdog setpoints.
    # the WD2 event can be reset with the command "OUT_WD2@0", which also stops the watchdog
    START_HEATER = "START_2"
    STOP_HEATER = "STOP_2"
    START_MOTOR = "START_4"
    STOP_MOTOR = "STOP_4"
    READ_HEATER_STATUS = "STATUS_2"  # Does not exist - find this
    READ_MOTOR_STATUS = "STATUS 4"  # Does not exist - find this
    READ_SOFTWARE_VERSION = "IN_VERSION"
    READ_SOFTWARE_ID = "IN_SOFTWARE_ID"  # Read software ID and version
    SET_ROTATION_CCW = "OUT_MODE_1"  # Sets rotation to CCW (OUT_MODE_2 does not set to CW)
    # Will not work if the last command sent to the shaker was RESET for some reason??
    RESET = "RESET"  # This will set rotation back to CW after being set to CCW


class Shaker(ShakerProtocol, IKADevice):
    """Driver for IKA orbital shaker."""

    def __init__(self, address, **kwargs):
        """Set up connection parameters, serial or IP address and port."""
        if address.startswith('/dev') or address.startswith('COM'):  # serial
            self.hw: Client = SerialClient(address=address, **kwargs)
        else:
            self.hw = TcpClient(address=address, **kwargs)
        self.lock = None  # needs to be initialized later, when the event loop exists

    async def get(self):
        """Get orbital shaker speed."""
        temp = await self.query(self.READ_ACTUAL_TEMPERATURE)
        temp_sp = await self.query(self.READ_SET_TEMPERATURE)
        heater_status = await self.query(self.READ_HEATER_STATUS)
        speed = await self.query(self.READ_ACTUAL_SPEED)
        speed_sp = await self.query(self.READ_SET_SPEED)
        shaker_status = await self.query(self.READ_MOTOR_STATUS)
        response = {
            'temp': {
                'setpoint': temp_sp,
                'actual': temp,
                'active': heater_status,
            },
            'speed': {
                'setpoint': int(speed_sp) if isinstance(speed_sp, float) else speed_sp,
                'actual': int(speed) if isinstance(speed, float) else speed,
                'active': shaker_status,
            }
        }
        return response

    async def get_info(self):
        """Get name and software version of orbital shaker."""
        name = await self.query(self.READ_DEVICE_NAME)
        version = await self.query(self.READ_SOFTWARE_VERSION)
        software_id = await self.query(self.READ_SOFTWARE_ID)
        response = {
            'name': name,
            'version': version,
            'software_ID': software_id,
        }
        return response

    async def set(self, equipment: str, setpoint: float):
        """Set a temperature or shaker speed setpoint."""
        if equipment == 'heater':
            if setpoint < 1.0 or setpoint > 100:
                raise ValueError('Setpoint invalid. Temperature SP must be between 1C and 100C.')
            await self.command(self.SET_TEMP + str(setpoint))
        elif equipment == 'shaker':
            if setpoint < 300 or setpoint > 3000:
                raise ValueError('Setpoint invalid. Speed SP must be between 300 and 3000rpm.')
            await self.command(self.SET_SPEED + str(setpoint))
        else:
            raise ValueError(f'Equipment "{equipment} invalid. '
                             'Must be either "heater" or "shaker"')

    async def control(self, equipment: str, on: bool):
        """Control the heater controlling process temperature, or shaker motor."""
        if equipment == 'heater':
            await self.command(self.START_HEATER if on else self.STOP_HEATER)
        elif equipment == 'shaker':
            await self.command(self.START_MOTOR if on else self.STOP_MOTOR)
        else:
            raise ValueError(f'Equipment "{equipment} invalid. '
                             'Must be either "heater" or "shaker"')

class VacuumProtocol:
    """Protocol for communicating with a vacuum pump.

    Command syntax and format from the manual:
        - commands and parameters are transmitted as capital letters
        - commands and parameters including successive parameters are separated by at least
          one space (hex 0x20)
        - each individual command (including parameters and data and each response are terminated
          with Blank CR LF (hex 0x20 hex 0x0d hex 0x0A) and have a maximum length of 80 characters
        - the decimal separator in a number is a dt (hex 0x2E)
    """

    # vacuum pump NAMUR commands
    READ_PARAMETERS = "IN_PARA1"
    SET_PARAMATERS_PUMP = "OUT_PARA1 "  # requires a value to be appended
    SET_PARAMETERS_BLUETOOTH = "OUT_PARA2 "  # requires a value to be appended
    READ_VAC_STATUS = "IN_STATUS"
    # Send the actual device status: 'OUT_STATUS'
    READ_SOFTWARE_VERSION = "IN_VERSION"
    # Read the release date of the display/ logic firmware: 'IN_DATE'
    READ_DEVICE_NAME = "IN_NAME"
    SET_DEVICE_NAME = "OUT_NAME "  # manual incorrectly refers to CUSTOM_DEVICE_NAME
    # Read the device type.: 'IN_DEVICE'
    # Read mac address of Wico.: 'IN_ADDRESS'
    # Read paired mac address of station.: 'IN_PARING' (sic)
    # Write new paired mac addresses of both station and Wico: 'OUT_ADDRESS'
    READ_SET_PRESSURE = "IN_SP_66"
    SET_PRESSURE = "OUT_SP_66 "  # requires a value to be appended
    READ_ACTUAL_PRESSURE = "IN_PV_66"
    READ_VAC_MODE = "IN_MODE_66"
    SET_VAC_MODE = "OUT_MODE_66 "  # requires a value to be appended
    # Reads error state: 'IN_ERROR'
    # Test Error. Sends out error code: 'OUT_ERROR'
    # Reads Bluetooth Device Name: 'IN_BT_NAME'
    # Reads custom device name: 'IN_CUSTOM_DEVICE_NAME'
    # Reads communication watchdog time: 'OUT_CUSTOM_DEVICE_NAME'
    # Sets communication watchdog time: 'IN_WD1@'
    # Set PC communication watchdog time 2: 'OUT_WD1@'
    # Set the PC safety pump rate: 'OUT_WD2@'
    # Set the PC safety pressure: 'OUT_SP_41'
    RESET = "RESET"
    START_MEASUREMENT = "START_66"
    STOP_MEASUREMENT = "STOP_66"
    # Starts IAP mode: 'ENTER_IAP'
    # It is used to calibrate vacuum: 'CALIB_66'
    # It is used to read vacuum calibration values: 'IN_CALIB_66'
    # It is used to calibrate vacuum: 'OUT_CALIB_66'

    ERROR_CODES = {
        3: ("The device temperature has exceeeded its limit."
            "Have you tried turning it off and on again?"),
        4: "The motor has overloaded.  Have you tried turning it off and on again?",
        8: "The speed sensor has faulted.  Contact service.",
        9: "The internal flash has a read or write error.  Contact service."
    }

    class Mode(Enum):
        """Possible operating modes."""

        AUTOMATIC = '0'  # Automatic boiling point recognition
        MANUAL = '1'  # Pressure control
        PERCENT = '2'  # % pump speed
        PROGRAM = '3'  # User-defined program


class Vacuum(VacuumProtocol, IKADevice):
    """Driver for IKA vacuum pump."""

    def __init__(self, address, **kwargs):
        """Set up connection parameters, serial or IP address and port."""
        if address.startswith('/dev') or address.startswith('COM'):  # serial
            self.hw: Client = SerialClient(address=address, **kwargs)
        else:
            self.hw = TcpClient(address=address, **kwargs)
        self.lock = None  # needs to be initialized later, when the event loop exists

    async def get_pressure(self) -> float:
        """Get vacuum pressure, converting to mmHg."""
        raw_pressure = await self.query(self.READ_ACTUAL_PRESSURE)
        return round(float(raw_pressure) / 1.333, 2)

    async def get_pressure_setpoint(self) -> float:
        """Get vacuum pressure setpoint, in mmHg."""
        raw_sp = await self.query(self.READ_SET_PRESSURE)
        return round(float(raw_sp), 2)

    async def get_status(self) -> bool:
        """Get vacuum status and convert to running/not running bool.

        TODO: Figure out the actual status bit (bit 25?)
        """
        raw_status = await self.query(self.READ_VAC_STATUS)
        return not (raw_status == '75' or raw_status == '79'
                    or raw_status == '203' or raw_status == '207'
                    or raw_status[0:2] == '32')

    async def get_vac_mode(self) -> str:
        """Get vacuum mode."""
        raw_mode = await self.query(self.READ_VAC_MODE)
        return raw_mode

    async def get(self) -> Dict[str, Any]:
        """Get pump operating data."""
        pressure = await self.get_pressure()
        pressure_sp = await self.get_pressure_setpoint()
        vac_mode = await self.get_vac_mode()
        vac_status = await self.get_status()

        response = {
            'active': vac_status,
            'mode': VacuumProtocol.Mode(vac_mode).name,
            'pressure': {
                'setpoint': pressure_sp,
                'actual': pressure,
            }
        }
        return response

    async def get_info(self) -> Dict[str, str]:
        """Get name and software version of vacuum."""
        name = await self.query(self.READ_DEVICE_NAME)
        version = await self.query(self.READ_SOFTWARE_VERSION)
        response = {
            'name': name,
            'version': version,
        }
        return response

    async def set(self, setpoint: float):
        """Set a vacuum pressure setpoint, converting from mmHg to mbar.

        Unlike other commands, the vacuum echoes back, so use query().
        """
        setpoint_mbar = str(int(setpoint * 1.333))
        await self.query(self.SET_PRESSURE + setpoint_mbar)

    async def set_mode(self, mode: VacuumProtocol.Mode):
        """Set the operating mode.

        Unlike other commands, the vacuum echoes back, so use query().
        """
        await self.query(self.SET_VAC_MODE + str(mode.value))

    async def set_name(self, name: str):
        """Set a custom device name.

        Unlike other commands, the vacuum echoes back, so use query().
        """
        await self.query(self.SET_DEVICE_NAME + name)

    async def control(self, on: bool):
        """Control the vacuum running status.

        Unlike other commands, the vacuum echoes back, so use query().
        """
        await self.query(self.START_MEASUREMENT if on else self.STOP_MEASUREMENT)



"""Test the overhead stirrer driver responds with correct data."""
from random import randint
from unittest import mock

import pytest

from ika import command_line
from ika.mock import OverheadStirrer

ADDRESS = '192.168.10.12:23'


@pytest.fixture
def driver():
    """Confirm the overhead stirrer correctly initializes."""
    return OverheadStirrer(ADDRESS)


@mock.patch('ika.OverheadStirrer', OverheadStirrer)
def test_driver_cli_with_info(capsys):
    """Confirm the commandline interface works."""
    command_line([ADDRESS, '-t', 'overhead'])
    captured = capsys.readouterr()
    assert "torque" in captured.out
    assert "name" in captured.out
    assert "null" not in captured.out


@mock.patch('ika.OverheadStirrer', OverheadStirrer)
def test_driver_cli(capsys):
    """Confirm the commandline interface works with --no-info."""
    command_line([ADDRESS, '-t', 'overhead', '--no-info'])
    captured = capsys.readouterr()
    assert "torque" in captured.out
    assert "name" not in captured.out
    assert "null" not in captured.out


async def test_get_response(driver):
    """Confirm that the driver returns correct values on get_info() calls."""
    info = await driver.get_info()
    assert 'IKA ES 60' in info['name'] or 'STIRR GO WHIRRR' in info['name']
    assert (isinstance(info['torque_limit'], float)
            and info['torque_limit'] <= 60.0 and info['torque_limit'] > 0.0)
    assert (isinstance(info['speed_limit'], float)
            and info['speed_limit'] <= 2000.0 and info['speed_limit'] > 30.0)


async def test_readme_example():
    """Confirm the readme example using an async context manager works."""
    async def get():
        async with OverheadStirrer(ADDRESS) as device:
            await device.get()       # Get speed, torque, temp
            info = await device.get_info()  # get name
            assert 'IKA ES 60' in info['name'] or 'STIRR GO WHIRRR' in info['name']
            assert (isinstance(info['torque_limit'], float)
                    and info['torque_limit'] <= 60.0 and info['torque_limit'] > 0.0)
            assert (isinstance(info['speed_limit'], float)
                    and info['speed_limit'] <= 2000.0 and info['speed_limit'] > 30.0)

    await get()


async def test_setpoint_roundtrip():
    """Confirm that setpoints can be updated."""
    async def get():
        async with OverheadStirrer(ADDRESS) as device:
            speed_limit = randint(50, 2000)
            speed_sp = randint(30, speed_limit)
            torque_limit = randint(5, 60)
            await device.control(on=True)
            await device.set(equipment='speed_limit', setpoint=speed_limit)
            await device.set(equipment='speed', setpoint=speed_sp)
            await device.set(equipment='torque_limit', setpoint=torque_limit)
            await device.control(on=False)
            response = await device.get()
            response.update(await device.get_info())
            assert speed_limit == response['speed_limit']
            assert speed_sp == response['speed']['setpoint']
            assert torque_limit == response['torque_limit']
    await get()


async def test_start_stop():
    """Confirm that the stirrer motor can be controlled."""
    async def get():
        async with OverheadStirrer(ADDRESS) as device:
            await device.control(on=True)
            response = await device.get()
            assert response['speed']['active'] is True

            await device.control(on=False)
            response = await device.get()
            assert response['speed']['active'] is False
    await get()


@pytest.mark.skip
async def test_reset():
    """Confirm that the reset functionality works."""
    async def get():
        async with OverheadStirrer(ADDRESS) as device:
            await device.reset()
            raise NotImplementedError
    await get()



"""Base functionality for async TCP communication.

Distributed under the GNU General Public License v3
Copyright (C) 2022 NuMat Technologies
"""
import asyncio
import logging
from abc import abstractmethod

import serial

# Add logger to module
logger = logging.getLogger('ika')


class Client:
    """Serial or TCP client."""

    def __init__(self, timeout):
        """Initialize common attributes."""
        self.address = ''
        self.open = False
        self.timeout = timeout
        self.timeouts = 0
        self.max_timeouts = 10
        self.connection = {}
        self.reconnecting = False
        self.eol = b'\r\n'

    @abstractmethod
    async def _write(self, message):
        """Write a message to the device."""
        pass

    @abstractmethod
    async def _read(self, length):
        """Read a fixed number of bytes from the device."""
        pass

    @abstractmethod
    async def _readline(self):
        """Read until a LF terminator."""
        pass

    async def _write_and_read(self, command):
        """Write a command and read a response."""
        await self._handle_connection()
        if self.open:
            try:
                response = await self._handle_communication(command)
                if response is None:
                    return None
                elif 'IN_VERSION' in command or 'IN_NAME' in command or 'TYPE' in command:
                    return response
                elif 'IN_PV_4' in response:
                    raise ConnectionError(
                        'Hotplate configured to communciate with a Eurostar overhead stirrer. '
                        'This must be turned off in the hotplate settings.'
                    )
                elif command in response:  # vacuum replies to queries by echoing the query
                    return response.strip(command).strip()
                elif command[-1] != response[-1]:
                    # all others reply to queries by echoing the query at the end
                    logger.error(f'Invalid response {response} to command {command}.')
                    await self._clear()
                    return None
                elif 'STATUS_2' in command:
                    raise NotImplementedError  # not sure how to interpret response of '-90 2'
                elif 'STATUS_4' in command:  # overhead stirrer
                    return response[0] == '1'
                elif 'START_4' in response or 'STOP_4' in response:  # overhead stirrer
                    return True
                elif 'STATUS' in command:  # hotplate
                    return response[0:2] == '11'  # undocumented, 11 = active, 12 = inactive
                return float(response[:-2])  # strip response command readback
            except asyncio.exceptions.IncompleteReadError:
                logger.error('IncompleteReadError.  Are there multiple connections?')
                return None
        else:
            return None

    async def _clear(self):
        """Clear the reader stream when it has been corrupted from multiple connections."""
        logger.warning("Multiple connections detected; clearing reader stream.")
        try:
            junk = await asyncio.wait_for(self._read(100), timeout=0.5)
            logger.warning(junk)
        except TimeoutError:
            pass

    async def _handle_communication(self, command):
        """Manage communication, including timeouts and logging."""
        try:
            await self._write(command)
            await asyncio.sleep(1)
            result = await self._readline()
            # result = await asyncio.wait_for(future, timeout=0.5)
            self.timeouts = 0
            return result
        except (asyncio.TimeoutError, TypeError, OSError):
            self.timeouts += 1
            if self.timeouts == self.max_timeouts:
                logger.error(f'Reading from {self.address} timed out '
                             f'{self.timeouts} times.')
                self.close()
            return None

    @abstractmethod
    def _handle_connection(self):
        pass

    @abstractmethod
    def close(self):
        """Close the connection."""
        pass


class TcpClient(Client):
    """A generic reconnecting asyncio TCP client.

    This base functionality can be used by any industrial control device
    communicating over TCP.
    """

    def __init__(self, address, timeout=1):
        """Communicator using a TCP/IP<=>serial gateway."""
        super().__init__(timeout)
        try:
            self.address, self.port = address.split(':')
        except ValueError:
            raise ValueError('address must be hostname:port')

    async def __aenter__(self):
        """Provide async entrance to context manager.

        Contrasting synchronous access, this will connect on initialization.
        """
        await self._handle_connection()
        return self

    def __exit__(self, *args):
        """Provide exit to context manager."""
        self.close()

    async def __aexit__(self, *args):
        """Provide async exit to context manager."""
        self.close()

    async def _connect(self):
        """Asynchronously open a TCP connection with the server."""
        self.close()
        reader, writer = await asyncio.open_connection(self.address, self.port)
        self.connection = {'reader': reader, 'writer': writer}
        self.open = True

    async def _read(self, length: int):
        """Read a fixed number of bytes from the device."""
        await self._handle_connection()
        response = await self.connection['reader'].read(length)
        return response.decode().strip()

    async def _readline(self):
        """Read until a line terminator."""
        await self._handle_connection()
        response = await self.connection['reader'].readuntil(self.eol)
        return response.decode().strip()

    async def _write(self, command: str):
        """Write a command and do not expect a response.

        As industrial devices are commonly unplugged, this has been expanded to
        handle recovering from disconnects.
        """
        await self._handle_connection()
        self.connection['writer'].write(command.encode() + self.eol)

    async def _handle_connection(self):
        """Automatically maintain TCP connection."""
        if self.open:
            return
        try:
            await asyncio.wait_for(self._connect(), timeout=0.75)
            self.reconnecting = False
        except (asyncio.TimeoutError, OSError):
            if not self.reconnecting:
                logger.error(f'Connecting to {self.address} timed out.')
            self.reconnecting = True

    async def _handle_communication(self, command):
        """Manage communication, including timeouts and logging."""
        try:
            await self._write(command)
            future = self._readline()
            result = await asyncio.wait_for(future, timeout=0.75)
            self.timeouts = 0
            return result
        except (asyncio.TimeoutError, TypeError, OSError):
            self.timeouts += 1
            if self.timeouts == self.max_timeouts:
                logger.error(f'Reading from {self.address} timed out '
                             f'{self.timeouts} times.')
                self.close()
            return None

    def close(self):
        """Close the TCP connection."""
        if self.open:
            self.connection['writer'].close()
        self.open = False


class SerialClient(Client):
    """Client using a directly-connected RS232 serial device."""

    def __init__(self, address=None, baudrate=9600, timeout=.15, bytesize=serial.SEVENBITS,
                 stopbits=serial.STOPBITS_ONE, parity=serial.PARITY_EVEN):
        """Initialize serial port."""
        super().__init__(timeout)
        self.address = address
        assert isinstance(self.address, str)
        self.serial_details = {'baudrate': baudrate,
                               'bytesize': bytesize,
                               'stopbits': stopbits,
                               'parity': parity,
                               'timeout': timeout}
        self.ser = serial.Serial(self.address, **self.serial_details)

    async def _read(self, length: int):
        """Read a fixed number of bytes from the device."""
        return self.ser.read(length).decode()

    async def _readline(self):
        """Read until a LF terminator."""
        return self.ser.readline().strip().decode()

    async def _write(self, message: str):
        """Write a message to the device."""
        self.ser.write(message.encode() + self.eol)

    def close(self):
        """Release resources."""
        self.ser.close()

    async def _handle_connection(self):
        self.open = True



"""Test the vacuum driver responds with correct data."""
from random import randint
from unittest import mock

import pytest

from ika import command_line
from ika.driver import VacuumProtocol
from ika.mock import Vacuum

ADDRESS = 'fakeip:123'


@pytest.fixture
def driver():
    """Confirm the vacuum correctly initializes."""
    return Vacuum(ADDRESS)


@pytest.fixture
def expected_response():
    """Return mocked vacuum data."""
    return {
        'name': "THIS SUCKS",
        'version': '1.3.001',
    }


@mock.patch('ika.Vacuum', Vacuum)
def test_driver_cli_with_info(capsys):
    """Confirm the commandline interface works."""
    command_line([ADDRESS, '--type', 'vacuum'])
    captured = capsys.readouterr()
    assert 'pressure' in captured.out
    assert 'name' in captured.out


@mock.patch('ika.Vacuum', Vacuum)
def test_driver_cli(capsys):
    """Confirm the commandline interface works with --no-info."""
    command_line([ADDRESS, '--type', 'vacuum', '--no-info'])
    captured = capsys.readouterr()
    assert 'pressure' in captured.out
    assert 'name' not in captured.out


async def test_get_response(driver, expected_response):
    """Confirm that the driver returns correct values on get_info() calls."""
    assert expected_response == await driver.get_info()


async def test_readme_example(expected_response):
    """Confirm the readme example using an async context manager works."""
    async def get():
        async with Vacuum(ADDRESS) as device:
            await device.get()
            assert expected_response == await device.get_info()  # Get name
    await get()


async def test_start_stop():
    """Confirm that the vacuum motor can be controlled."""
    async def get():
        async with Vacuum(ADDRESS) as device:
            await device.control(on=True)
            response = await device.get()
            assert response['active'] is True

            await device.control(on=False)
            response = await device.get()
            assert response['active'] is False
    await get()


async def test_setpoint_roundtrip():
    """Confirm that the pressure setpoint can be updated."""
    async def get():
        async with Vacuum(ADDRESS) as device:
            await device.control(on=True)
            pressure_sp = randint(0, 760)
            await device.set(setpoint=pressure_sp)
            response = await device.get()
            assert pressure_sp == pytest.approx(response['pressure']['setpoint'], 2)
            await device.control(on=False)
    await get()


@pytest.mark.parametrize('mode', list(VacuumProtocol.Mode))
async def test_mode_roundtrip(mode):
    """Confirm that the various vacuum modes can be updated."""
    async def get():
        async with Vacuum(ADDRESS) as device:
            await device.set_mode(mode)
            response = await device.get()
            assert mode.name == response['mode']
    await get()


async def test_name_roundtrip():
    """Confirm that the device name can be updated."""
    async def get():
        async with Vacuum(ADDRESS) as device:
            await device.set_name('THIS SUCKS')
            name = (await device.get_info())['name']
            assert name == 'THIS SUCKS'
            await device.set_name('VACSTAR control')
            name = (await device.get_info())['name']
            assert name == 'VACSTAR control'
    await get()



"""Python driver and command line tool for IKA instruments."""
from setuptools import setup

with open('README.md') as in_file:
    long_description = in_file.read()

setup(
    name="ika-control",
    version="0.6.1",
    description="Python driver for IKA instruments.",
    long_description=long_description,
    long_description_content_type='text/markdown',
    url="https://github.com/numat/ika/",
    author="Alex Ruddick",
    author_email="a.ruddick@numat-tech.com",
    packages=['ika'],
    install_requires=['pyserial'],
    extras_require={
        'test': [
            'pytest>=6,<8',
            'pytest-cov>=5,<6',
            'pytest-asyncio==0.*',
            'pytest-xdist==3.*',
            'ruff==0.4.7',
            'mypy==1.10.1',
            'types-pyserial==3.5.0.20240311'
        ],
    },
    entry_points={
        'console_scripts': [('ika = ika:command_line')]
    },
    license='GPLv3',
    classifiers=[
        'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
        'Development Status :: 4 - Beta',
        'Natural Language :: English',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Programming Language :: Python :: 3.11',
        'Programming Language :: Python :: 3.12',
        'Topic :: Scientific/Engineering :: Human Machine Interfaces',
    ],
)



"""
Python driver for IKA equipment.

Distributed under the GNU General Public License v3
Copyright (C) 2022 NuMat Technologies
"""
from ika.driver import Hotplate, OverheadStirrer, Shaker, Vacuum


def command_line(args=None):
    """Command line tool exposed through package install."""
    import argparse
    import asyncio
    import json

    parser = argparse.ArgumentParser(description="Read device status.")
    parser.add_argument('address', type=str, help="The target TCP address:port")
    parser.add_argument('-t', '--type', help="The type of device (default 'overhead')",
                        type=str, default='overhead')
    parser.add_argument('-n', '--no-info', action='store_true', help="Exclude "
                        "device information. Reduces communication overhead.")
    args = parser.parse_args(args)
    if args.type == 'overhead':
        async def get():
            async with OverheadStirrer(args.address) as device:
                d = await device.get()
                if not args.no_info:
                    d['info'] = await device.get_info()
                print(json.dumps(d, indent=4))
    elif args.type == 'hotplate':
        async def get():
            async with Hotplate(args.address,
                                include_surface_control=not args.no_info) as device:
                d = await device.get()
                if not args.no_info:
                    d['info'] = await device.get_info()
                print(json.dumps(d, indent=4))
    elif args.type == 'shaker':
        async def get():
            async with Shaker(args.address) as device:
                d = await device.get()
                if not args.no_info:
                    d['info'] = await device.get_info()
                print(json.dumps(d, indent=4))
    elif args.type == 'vacuum':
        async def get():
            async with Vacuum(args.address) as device:
                d = await device.get()
                if not args.no_info:
                    d['info'] = await device.get_info()
                print(json.dumps(d, indent=4))
    else:
        raise ValueError(f"Unsupported device type: {args.type}")
    asyncio.run(get())


if __name__ == '__main__':
    command_line()
